import logging

import pendulum

from dbnd._core.utils.timezone import utcnow
from dbnd_airflow_export.dag_processing import get_dags, load_dags_models
from dbnd_airflow_export.plugin_old.db_queries import (
    get_completed_task_instances_and_dag_runs,
    get_dag_runs_within_time_window,
    get_dag_runs_without_end_date,
    get_incomplete_task_instances_from_completed_dag_runs,
    get_task_instances_without_end_date,
)
from dbnd_airflow_export.plugin_old.metrics import measure_time
from dbnd_airflow_export.plugin_old.model import ExportData


def get_dags_list_only(session, dagbag, dag_ids):
    """
    This function returns all dags from Airflow but in their raw form - No tasks or source code are attached.
    Please do not use it in cases where either tasks or source code are required.
    """
    load_dags_models(session)
    dags_list = get_dags(
        dagbag=dagbag, include_task_args=False, dag_ids=dag_ids, raw_data_only=True
    )
    ed = ExportData(
        task_instances=[], dag_runs=[], dags=dags_list, since=str(utcnow()),
    )

    return ed


@measure_time
def get_complete_data(
    since,
    dag_ids,
    dagbag,
    quantity,
    include_logs,
    include_task_args,
    include_xcom,
    session,
):
    """
    This function returns the following data:
    1. The first task instances that ended after since (limited by quantity). Also get their dag runs.
    2. Dag runs that don't have task instances, but whose end_date is after since and before end_date of the most
    recent of the task instances.
    3. All Dags of the Dag runs that we found in steps 1 and 2
    """
    load_dags_models(session)

    logging.info("Trying to query completed task instances and dagruns from %s" % since)

    task_instances, dag_runs = get_completed_task_instances_and_dag_runs(
        since, dag_ids, quantity, dagbag, include_logs, include_xcom, session
    )
    logging.info("%d task instances were found." % len(task_instances))

    task_end_dates = [
        task.end_date for task in task_instances if task.end_date is not None
    ]
    if not task_end_dates or not quantity or len(task_instances) < quantity:
        dag_run_end_date = pendulum.datetime.max
    else:
        dag_run_end_date = max(task_end_dates)

    dag_runs |= get_dag_runs_within_time_window(
        since, dag_run_end_date, dag_ids, quantity, session
    )
    logging.info("%d dag runs were found." % len(dag_runs))

    if not dag_ids:
        dag_ids = set(dag_run.dag_id for dag_run in dag_runs)

    dags_list = get_dags(
        dagbag=dagbag,
        include_task_args=include_task_args,
        dag_ids=dag_ids,
        raw_data_only=False,
    )

    logging.info(
        "Returning {} task instances, {} dag runs, {} dags".format(
            len(task_instances), len(dag_runs), len(dags_list)
        )
    )

    ed = ExportData(
        task_instances=task_instances, dag_runs=dag_runs, dags=dags_list, since=since,
    )

    return ed


@measure_time
def get_incomplete_data_type_1(
    since, dag_ids, dagbag, max_quantity, include_task_args, offset, session
):
    """
    This function returns the following data:
    1. Task instances with end_date=None from dag runs that finished running (their end_date is not None).
    It brings only the first dag runs (limited by max_quantity) that ended after since.
    2. All Dags of the Dag runs that we found in step 1
    """
    load_dags_models(session)

    logging.info(
        "Trying to query incomplete task instances from complete dagruns from %s"
        % since
    )

    task_instances, dag_runs = get_incomplete_task_instances_from_completed_dag_runs(
        since, dag_ids, dagbag, max_quantity, offset, session
    )

    if not dag_ids:
        dag_ids = set(dag_run.dag_id for dag_run in dag_runs)

    dags_list = get_dags(
        dagbag=dagbag,
        include_task_args=include_task_args,
        dag_ids=dag_ids,
        raw_data_only=False,
    )

    logging.info(
        "Returning {} task instances, {} dag runs, {} dags".format(
            len(task_instances), len(dag_runs), len(dags_list)
        )
    )

    ed = ExportData(
        task_instances=task_instances, dag_runs=dag_runs, dags=dags_list, since=since,
    )

    return ed


@measure_time
def get_incomplete_data_type_2(
    since, dag_ids, dagbag, quantity, include_task_args, incomplete_offset, session
):
    """
    This function returns the following data:
    1. Task instances with end_date=None from dag runs that are running (their end_date=None). Also their dag runs.
    2. All Dags of the Dag runs that we found in step 1
    Important: since pagination is used, the same parameters can return different results at different times.
    """
    load_dags_models(session)

    logging.info(
        "Trying to query incomplete task instances and dagruns from %s" % since
    )

    task_instances, dag_runs = get_task_instances_without_end_date(
        since=since,
        dag_ids=dag_ids,
        dagbag=dagbag,
        session=session,
        incomplete_offset=incomplete_offset,
        page_size=quantity,
    )
    logging.info(
        "Found {} task instances with no end_date from {} dag runs".format(
            len(task_instances), len(dag_runs)
        )
    )

    dag_runs_without_date = get_dag_runs_without_end_date(
        since=since,
        dag_ids=dag_ids,
        session=session,
        incomplete_offset=incomplete_offset,
        page_size=quantity,
    )
    logging.info("Found {} dag runs with no end_date".format(len(dag_runs)))

    dag_runs |= dag_runs_without_date

    if not dag_ids:
        dag_ids = set(dag_run.dag_id for dag_run in dag_runs)

    dags_list = get_dags(
        dagbag=dagbag,
        include_task_args=include_task_args,
        dag_ids=dag_ids,
        raw_data_only=False,
    )

    logging.info(
        "Returning {} task instances, {} dag runs, {} dags".format(
            len(task_instances), len(dag_runs), len(dags_list)
        )
    )

    ed = ExportData(
        task_instances=task_instances, dag_runs=dag_runs, dags=dags_list, since=since,
    )

    return ed



from airflow.models import DagModel
from airflow.utils.db import provide_session

from dbnd_airflow_export.dag_processing import (
    get_current_dag_model,
    get_dags,
    load_dags_models,
)
from dbnd_airflow_export.models import (
    AirflowNewDagRun,
    DagRunsStatesData,
    FullRunsData,
    LastSeenData,
    NewRunsData,
)
from dbnd_airflow_export.queries import (
    find_all_logs_grouped_by_runs,
    find_full_dag_runs,
    find_max_dag_run_id,
    find_max_log_run_id,
    find_new_dag_runs,
)


@provide_session
def get_last_seen_values(session=None):
    max_dag_run_id = find_max_dag_run_id(session)
    max_log_id = find_max_log_run_id(session)

    return LastSeenData(
        last_seen_dag_run_id=max_dag_run_id, last_seen_log_id=max_log_id
    )


@provide_session
def get_new_dag_runs(
    last_seen_dag_run_id, last_seen_log_id, extra_dag_runs_ids, session=None
):
    max_dag_run_id = find_max_dag_run_id(session)
    max_log_id = find_max_log_run_id(session)

    if last_seen_dag_run_id is None:
        last_seen_dag_run_id = max_dag_run_id

    if last_seen_log_id is None:
        last_seen_log_id = max_log_id

    logs = find_all_logs_grouped_by_runs(last_seen_log_id, session)
    logs_dict = {(log.dag_id, log.execution_date): log for log in logs}

    dag_runs = find_new_dag_runs(
        last_seen_dag_run_id, extra_dag_runs_ids, logs_dict.keys(), session
    )

    new_dag_runs = []
    for dag_run in dag_runs:
        log = logs_dict.get((dag_run.dag_id, dag_run.execution_date), None)

        if log is None:
            events = []
        elif isinstance(log.events, str):  # mysql, sqlite
            events = log.events.split(",")
        else:  # postgres
            events = log.events

        new_dag_run = AirflowNewDagRun(
            id=dag_run.id,
            dag_id=dag_run.dag_id,
            execution_date=dag_run.execution_date,
            state=dag_run.state,
            is_paused=dag_run.is_paused,
            has_updated_task_instances=(dag_run.dag_id, dag_run.execution_date)
            in logs_dict,
            events=events,
            max_log_id=log.id if log else None,
        )
        new_dag_runs.append(new_dag_run)

    new_runs = NewRunsData(
        new_dag_runs=new_dag_runs,
        last_seen_dag_run_id=max_dag_run_id,
        last_seen_log_id=max_log_id,
    )
    return new_runs


@provide_session
def get_full_dag_runs(dag_run_ids, airflow_dagbag=None, session=None):
    if airflow_dagbag:
        dagbag = airflow_dagbag
    else:
        from airflow import settings

        if settings.RBAC:
            from airflow.www_rbac.views import dagbag
        else:
            from airflow.www.views import dagbag

    old_get_current_dag = DagModel.get_current
    try:
        DagModel.get_current = get_current_dag_model
        load_dags_models(session)
        task_instances, dag_runs = find_full_dag_runs(dag_run_ids, session)
        dag_ids = set(run.dag_id for run in dag_runs)
        dags = get_dags(dagbag, True, dag_ids, False)
        full_runs = FullRunsData(
            task_instances=task_instances, dag_runs=dag_runs, dags=dags
        )
        return full_runs
    finally:
        DagModel.get_current = old_get_current_dag


@provide_session
def get_dag_runs_states_data(dag_run_ids, session=None):
    task_instances, dag_runs = find_full_dag_runs(dag_run_ids, session)
    dag_runs_states_data = DagRunsStatesData(
        task_instances=task_instances, dag_runs=dag_runs
    )
    return dag_runs_states_data



import logging

import flask_admin
import flask_appbuilder

from airflow.plugins_manager import AirflowPlugin

from dbnd_airflow_export.plugin_old.data_exporting import export_data_api
from dbnd_airflow_export.request_processing import (
    process_dag_run_states_data_request,
    process_full_runs_request,
    process_last_seen_values_request,
    process_new_runs_request,
)


class ExportDataViewAppBuilder(flask_appbuilder.BaseView):
    endpoint = "data_export_plugin"
    default_view = "export_data"

    @flask_appbuilder.has_access
    @flask_appbuilder.expose("/export_data")
    def export_data(self):
        from airflow.www_rbac.views import dagbag

        return export_data_api(dagbag)

    @flask_appbuilder.has_access
    @flask_appbuilder.expose("/last_seen_values")
    def last_seen_values(self):
        return process_last_seen_values_request()

    @flask_appbuilder.has_access
    @flask_appbuilder.expose("/new_runs")
    def new_runs(self):
        return process_new_runs_request()

    @flask_appbuilder.has_access
    @flask_appbuilder.expose("/full_runs")
    def full_runs(self):
        return process_full_runs_request()

    @flask_appbuilder.has_access
    @flask_appbuilder.expose("/runs_states_data")
    def task_instances(self):
        return process_dag_run_states_data_request()


class ExportDataViewAdmin(flask_admin.BaseView):
    def __init__(self, *args, **kwargs):
        super(ExportDataViewAdmin, self).__init__(*args, **kwargs)
        self.endpoint = "data_export_plugin"
        self.default_view = "export_data"

    @flask_admin.expose("/")
    @flask_admin.expose("/export_data")
    def export_data(self):
        from airflow.www.views import dagbag

        return export_data_api(dagbag)

    @flask_admin.expose("/")
    @flask_admin.expose("/last_seen_values")
    def last_seen_values(self):
        return process_last_seen_values_request()

    @flask_admin.expose("/")
    @flask_admin.expose("/new_runs")
    def new_runs(self):
        return process_new_runs_request()

    @flask_admin.expose("/")
    @flask_admin.expose("/full_runs")
    def full_runs(self):
        return process_full_runs_request()

    @flask_admin.expose("/")
    @flask_admin.expose("/runs_states_data")
    def task_instances(self):
        return process_dag_run_states_data_request()


class DataExportAirflowPlugin(AirflowPlugin):
    name = "dbnd_airflow_export"
    admin_views = [ExportDataViewAdmin(category="Admin", name="Export Data")]
    appbuilder_views = [
        {"category": "Admin", "name": "Export Data", "view": ExportDataViewAppBuilder()}
    ]


try:
    # this import is critical for loading `requires_authentication`
    from airflow import api

    api.load_auth()

    from airflow.www_rbac.api.experimental.endpoints import (
        api_experimental,
        requires_authentication,
    )

    @api_experimental.route("/export_data", methods=["GET"])
    @requires_authentication
    def export_data():
        from airflow.www_rbac.views import dagbag

        return export_data_api(dagbag)

    @api_experimental.route("/last_seen_values", methods=["GET"])
    @requires_authentication
    def last_seen_values(self):
        return process_last_seen_values_request()

    @api_experimental.route("/new_runs", methods=["GET"])
    @requires_authentication
    def new_runs(self):
        return process_new_runs_request()

    @api_experimental.route("/full_runs", methods=["GET"])
    @requires_authentication
    def full_runs(self):
        return process_full_runs_request()

    @api_experimental.route("/runs_states_data", methods=["GET"])
    @requires_authentication
    def task_instances(self):
        return process_dag_run_states_data_request()


except Exception as e:
    logging.error("Export data could not be added to experimental api: %s", e)



import datetime
import json

from flask import Response


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        # convert dates and numpy objects in a json serializable format
        if isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y-%m-%d")

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def json_response(obj):
    return Response(
        response=json.dumps(obj, indent=4, cls=JsonEncoder),
        status=200,
        mimetype="application/json",
    )



from typing import List, Union

from airflow.models import DagModel, DagRun, Log, TaskInstance
from airflow.settings import Session
from sqlalchemy import and_, func, or_, tuple_

from dbnd_airflow_export.models import AirflowTaskInstance
from dbnd_airflow_export.plugin_old.metrics import measure_time, save_result_size
from dbnd_airflow_export.plugin_old.model import EDagRun


@save_result_size("find_new_dag_runs")
@measure_time
def find_new_dag_runs(
    last_seen_dagrun_id, extra_dag_runs_ids, extra_dag_runs_tuple, session
):
    new_runs_query = session.query(
        DagRun.id,
        DagRun.dag_id,
        DagRun.execution_date,
        DagRun.state,
        DagModel.is_paused,
    ).join(DagModel, DagModel.dag_id == DagRun.dag_id)

    new_runs_filter_condition = or_(
        DagRun.id.in_(extra_dag_runs_ids),
        and_(DagRun.state == "running", DagModel.is_paused.is_(False)),
        tuple_(DagRun.dag_id, DagRun.execution_date).in_(extra_dag_runs_tuple),
    )

    if last_seen_dagrun_id is not None:
        new_runs_filter_condition = or_(
            new_runs_filter_condition, DagRun.id > last_seen_dagrun_id
        )

    new_runs_query = new_runs_query.filter(new_runs_filter_condition)

    new_runs = new_runs_query.all()

    return new_runs


@save_result_size("find_all_logs_grouped_by_runs")
@measure_time
def find_all_logs_grouped_by_runs(last_seen_log_id, session):
    # type: (int, Session) -> List[Log]

    if last_seen_log_id is None:
        return []

    if session.bind.dialect.name == "postgresql":
        events_field = func.array_agg(Log.event.distinct()).label("events")
    else:  # mysql, sqlite
        events_field = func.group_concat(Log.event.distinct()).label("events")

    logs_query = (
        session.query(
            func.max(Log.id).label("id"), Log.dag_id, Log.execution_date, events_field
        )
        .filter(and_(Log.id > last_seen_log_id, Log.dag_id.isnot(None)))
        .group_by(Log.dag_id, Log.execution_date)
    )

    logs = logs_query.all()

    return logs


@measure_time
def find_max_dag_run_id(session):
    # type: (Session) -> Union[int, None]
    last_dag_run_id = session.query(func.max(DagRun.id)).scalar()
    return last_dag_run_id


@measure_time
def find_max_log_run_id(session):
    # type: (Session) -> Union[int, None]
    last_log_id = session.query(func.max(Log.id)).scalar()
    return last_log_id


@save_result_size("find_full_dag_runs")
@measure_time
def find_full_dag_runs(dag_run_ids, session):
    # type: (List[int], Session) -> (List[AirflowTaskInstance], List[EDagRun])
    result_fields = (
        session.query(*EDagRun.query_fields(), *AirflowTaskInstance.query_fields())
        .outerjoin(
            TaskInstance,
            (
                (TaskInstance.dag_id == DagRun.dag_id)
                & (TaskInstance.execution_date == DagRun.execution_date)
                | (TaskInstance.task_id.is_(None))
            ),
        )
        .filter(DagRun.id.in_(dag_run_ids))
        .all()
    )

    task_instances = []
    dag_runs = set()
    for fields in result_fields:
        dag_run_fields = fields[: len(EDagRun.db_fields)]
        dag_run = EDagRun.from_db_fields(*dag_run_fields)
        dag_runs.add(dag_run)

        task_instance_fields = fields[len(EDagRun.db_fields) :]
        if any(task_instance_fields):
            task_instance = AirflowTaskInstance(*task_instance_fields)
            task_instances.append(task_instance)

    return task_instances, dag_runs



import datetime
import logging
import re
import sys
import traceback

import flask
import pendulum

from airflow.configuration import conf
from airflow.models import DagModel
from airflow.utils.db import provide_session

from dbnd_airflow_export.dag_processing import get_current_dag_model
from dbnd_airflow_export.plugin_old.logic import (
    get_complete_data,
    get_dags_list_only,
    get_incomplete_data_type_1,
    get_incomplete_data_type_2,
)
from dbnd_airflow_export.utils import json_response


@provide_session
def get_airflow_data(
    dagbag,
    since,
    include_logs,
    include_task_args,
    include_xcom,
    fetch_type,
    dag_ids=None,
    quantity=None,
    incomplete_offset=None,
    session=None,
):
    if since:
        since = pendulum.parse(re.sub(r" 00:00$", "Z", str(since)))
    else:
        since = pendulum.datetime.min

    # We monkey patch `get_current` to optimize sql querying
    old_get_current_dag = DagModel.get_current
    try:
        DagModel.get_current = get_current_dag_model
        if fetch_type == "dags_only":
            result = get_dags_list_only(session, dagbag, dag_ids)
        elif fetch_type == "incomplete_type1":
            result = get_incomplete_data_type_1(
                since,
                dag_ids,
                dagbag,
                quantity,
                include_task_args,
                incomplete_offset,
                session,
            )
        elif fetch_type == "incomplete_type2":
            result = get_incomplete_data_type_2(
                since,
                dag_ids,
                dagbag,
                quantity,
                include_task_args,
                incomplete_offset,
                session,
            )
        else:
            result = get_complete_data(
                since,
                dag_ids,
                dagbag,
                quantity,
                include_logs,
                include_task_args,
                include_xcom,
                session,
            )
    finally:
        DagModel.get_current = old_get_current_dag

    if result:
        result = result.as_dict()

    return result


def export_data_api(dagbag):
    since = flask.request.args.get("since")
    include_logs = flask.request.args.get("include_logs")
    include_task_args = bool(flask.request.args.get("include_task_args"))
    include_xcom = bool(flask.request.args.get("include_xcom"))
    dag_ids = (
        flask.request.args.getlist("dag_ids")
        if "dag_ids" in flask.request.args
        else None
    )
    quantity = flask.request.args.get("fetch_quantity", type=int)
    rbac_enabled = conf.get("webserver", "rbac").lower() == "true"
    incomplete_offset = flask.request.args.get("incomplete_offset", type=int)
    fetch_type = flask.request.args.get("fetch_type")

    if not since and not include_logs and not dag_ids and not quantity:
        new_since = datetime.datetime.utcnow().replace(
            tzinfo=pendulum.timezone("UTC")
        ) - datetime.timedelta(days=1)
        redirect_url = (
            "ExportDataViewAppBuilder" if rbac_enabled else "data_export_plugin"
        )
        redirect_url += ".export_data"
        return flask.redirect(flask.url_for(redirect_url, since=new_since, code=303))

    # do_update = flask.request.args.get("do_update", "").lower() == "true"
    # verbose = flask.request.args.get("verbose", str(not do_update)).lower() == "true"

    try:
        export_data = get_airflow_data(
            dagbag=dagbag,
            since=since,
            include_logs=include_logs,
            include_task_args=include_task_args,
            include_xcom=include_xcom,
            dag_ids=dag_ids,
            quantity=quantity,
            fetch_type=fetch_type,
            incomplete_offset=incomplete_offset,
        )
        export_data["metrics"] = {
            "performance": flask.g.perf_metrics,
            "sizes": flask.g.size_metrics,
        }
        logging.info("Performance metrics %s", flask.g.perf_metrics)
    except Exception:
        exception_type, exception, exc_traceback = sys.exc_info()
        message = "".join(traceback.format_tb(exc_traceback))
        message += "{}: {}. ".format(exception_type.__name__, exception)
        logging.error("Exception during data export: \n%s", message)
        export_data = {"error": message}
    return json_response(export_data)



import logging

import flask

from airflow.version import version as airflow_version

from dbnd_airflow_export.api_functions import (
    get_dag_runs_states_data,
    get_full_dag_runs,
    get_last_seen_values,
    get_new_dag_runs,
)
from dbnd_airflow_export.models import AirflowExportData, AirflowExportMeta
from dbnd_airflow_export.utils import json_response


def process_last_seen_values_request():
    result = AirflowExportData()
    try:
        result = get_last_seen_values()
    except Exception as e:
        result = AirflowExportData()
        logging.error("Exception during data export: \n%s", str(e))
        result.error_message = str(e)
    finally:
        result.airflow_export_meta = get_meta()
        return json_response(result.as_dict())


def process_new_runs_request():
    last_seen_dag_run_id = flask.request.args.get("last_seen_dag_run_id", type=int)
    last_seen_log_id = flask.request.args.get("last_seen_log_id", type=int)
    extra_dag_runs_ids = (
        list(
            map(int, flask.request.args.get("extra_dag_runs_ids", type=str).split(","),)
        )
        if "extra_dag_runs_ids" in flask.request.args
        else []
    )

    result = AirflowExportData()
    try:
        result = get_new_dag_runs(
            last_seen_dag_run_id, last_seen_log_id, extra_dag_runs_ids
        )
    except Exception as e:
        result = AirflowExportData()
        logging.error("Exception during data export: \n%s", str(e))
        result.error_message = str(e)
    finally:
        result.airflow_export_meta = get_meta()
        return json_response(result.as_dict())


def process_full_runs_request():
    dag_run_ids = (
        list(map(int, flask.request.args.get("dag_run_ids", type=str).split(",")))
        if "dag_run_ids" in flask.request.args
        else []
    )

    result = AirflowExportData()
    try:
        result = get_full_dag_runs(dag_run_ids)
    except Exception as e:
        result = AirflowExportData()
        logging.error("Exception during data export: \n%s", str(e))
        result.error_message = str(e)
    finally:
        result.airflow_export_meta = get_meta()
        return json_response(result.as_dict())


def process_dag_run_states_data_request():
    dag_run_ids = (
        list(map(int, flask.request.args.get("dag_run_ids", type=str).split(",")))
        if "dag_run_ids" in flask.request.args
        else []
    )

    result = AirflowExportData()
    try:
        result = get_dag_runs_states_data(dag_run_ids)
    except Exception as e:
        result = AirflowExportData()
        logging.error("Exception during data export: \n%s", str(e))
        result.error_message = str(e)
    finally:
        result.airflow_export_meta = get_meta()
        return json_response(result.as_dict())


def get_meta():
    meta = AirflowExportMeta()
    meta.airflow_version = airflow_version
    meta.plugin_version = "2.0"
    meta.request_args = flask.request.args

    metrics = {}
    if hasattr(flask.g, "perf_metrics"):
        metrics["performance"] = flask.g.perf_metrics
    if hasattr(flask.g, "sizes"):
        metrics["sizes"]: flask.g.size_metrics

    meta.metrics = metrics

    return meta



import itertools
import logging
import sys

import six

from airflow.models import XCom

from dbnd_airflow_export.plugin_old.helpers import _get_log
from dbnd_airflow_export.plugin_old.metrics import measure_time
from dbnd_airflow_export.plugin_old.model import ETaskInstance


try:
    # in dbnd it might be overridden
    from airflow.models import original_TaskInstance as TaskInstance
except Exception:
    from airflow.models import TaskInstance


MAX_XCOM_SIZE_IN_BYTES = 10000
MAX_XCOM_LENGTH = 10


def build_task_instance(ti_fields, dagbag, include_xcom, include_logs, session):
    eti = ETaskInstance(*ti_fields)
    if include_xcom or include_logs:
        dag_from_dag_bag = dagbag.get_dag(eti.dag_id)
        if dag_from_dag_bag:
            if include_xcom:
                xcom_dict = _get_task_instance_xcom_dict(
                    dag_id=eti.dag_id,
                    task_id=eti.task_id,
                    execution_date=eti.execution_date,
                )
                eti.xcom_dict = xcom_dict
            if include_logs:
                task = (
                    dag_from_dag_bag.get_task(eti.task_id)
                    if dag_from_dag_bag and dag_from_dag_bag.has_task(eti.task_id)
                    else None
                )
                if task:
                    ti = (
                        session.query(TaskInstance)
                        .filter(
                            TaskInstance.task_id == eti.task_id,
                            TaskInstance.dag_id == eti.dag_id,
                            TaskInstance.execution_date == eti.execution_date,
                        )
                        .first()
                    )
                    eti.log_body = _get_log(ti, task)
    return eti


@measure_time
def _get_task_instance_xcom_dict(dag_id, task_id, execution_date):
    try:
        results = XCom.get_many(
            dag_ids=dag_id, task_ids=task_id, execution_date=execution_date
        )
        if not results:
            return {}

        xcom_dict = {xcom.key: str(xcom.value) for xcom in results}

        sliced_xcom = (
            dict(itertools.islice(xcom_dict.items(), MAX_XCOM_LENGTH))
            if len(xcom_dict) > MAX_XCOM_LENGTH
            else xcom_dict
        )
        for key, value in six.iteritems(sliced_xcom):
            sliced_xcom[key] = _shorten_xcom_value(value)

        return sliced_xcom
    except Exception as e:
        logging.info("Failed to get xcom dict. Exception: {}".format(e))
        return {}


def _shorten_xcom_value(xcom_value):
    if sys.getsizeof(xcom_value) <= MAX_XCOM_SIZE_IN_BYTES:
        return xcom_value

    diff = len(xcom_value) - MAX_XCOM_SIZE_IN_BYTES
    return xcom_value[-diff:]



import pendulum

from airflow.models import DagRun
from sqlalchemy import and_

from dbnd_airflow_export.plugin_old.metrics import measure_time, save_result_size
from dbnd_airflow_export.plugin_old.model import EDagRun, ETaskInstance
from dbnd_airflow_export.plugin_old.task_instance_building import build_task_instance


try:
    # in dbnd it might be overridden
    from airflow.models import original_TaskInstance as TaskInstance
except Exception:
    from airflow.models import TaskInstance


@save_result_size("dag_runs_without_date")
@measure_time
def get_dag_runs_without_end_date(
    since, dag_ids, session, page_size=100, incomplete_offset=0
):
    dagruns_query = session.query(*EDagRun.query_fields()).filter(
        and_(DagRun.end_date.is_(None)), DagRun.execution_date > since
    )

    if dag_ids:
        dagruns_query = dagruns_query.filter(DagRun.dag_id.in_(dag_ids))

    dagruns_query = (
        dagruns_query.order_by(DagRun.id).limit(page_size).offset(incomplete_offset)
    )

    return set(EDagRun.from_db_fields(*fields) for fields in dagruns_query.all())


@save_result_size("get_dag_runs_within_time_window")
@measure_time
def get_dag_runs_within_time_window(start_date, end_date, dag_ids, quantity, session):
    # Bring all dag runs with no tasks (limit the number)
    dagruns_query = session.query(*EDagRun.query_fields()).filter(
        and_(DagRun.end_date > start_date, DagRun.end_date <= end_date)
    )

    if dag_ids:
        dagruns_query = dagruns_query.filter(DagRun.dag_id.in_(dag_ids))

    # We reached a point where there are no more task, but can have potentially large number of dag runs with no tasks
    # so let's limit them. In the next fetch we'll get the next runs.
    if quantity is not None and end_date == pendulum.datetime.max:
        dagruns_query = dagruns_query.order_by(DagRun.end_date).limit(quantity)

    return set(EDagRun.from_db_fields(*fields) for fields in dagruns_query.all())


@save_result_size("task_instances_without_date")
@measure_time
def get_task_instances_without_end_date(
    since, dag_ids, dagbag, session, page_size=100, incomplete_offset=0
):
    task_instances_query = (
        session.query(*ETaskInstance.query_fields(), *EDagRun.query_fields())
        .join(
            DagRun,
            (TaskInstance.dag_id == DagRun.dag_id)
            & (TaskInstance.execution_date == DagRun.execution_date),
        )
        .filter(
            and_(
                DagRun.end_date.is_(None),
                TaskInstance.end_date.is_(None),
                TaskInstance.execution_date > since,
            )
        )
    )

    if dag_ids:
        task_instances_query = task_instances_query.filter(
            TaskInstance.dag_id.in_(dag_ids)
        )

    task_instances_query = (
        task_instances_query.order_by(
            TaskInstance.task_id, TaskInstance.dag_id, TaskInstance.execution_date
        )
        .limit(page_size)
        .offset(incomplete_offset)
    )

    results = task_instances_query.all()

    task_instances, dag_runs = _create_task_instances_and_dag_runs(
        results, dagbag, False, False, session
    )

    return task_instances, dag_runs


@save_result_size("completed_task_instances", "completed_dag_runs")
@measure_time
def get_completed_task_instances_and_dag_runs(
    since, dag_ids, quantity, dagbag, include_logs, include_xcom, session
):
    task_instances_query = (
        session.query(*ETaskInstance.query_fields(), *EDagRun.query_fields())
        .join(
            DagRun,
            (TaskInstance.dag_id == DagRun.dag_id)
            & (TaskInstance.execution_date == DagRun.execution_date),
        )
        .filter(TaskInstance.end_date > since)
    )

    if dag_ids:
        task_instances_query = task_instances_query.filter(
            TaskInstance.dag_id.in_(dag_ids)
        )

    if quantity is not None:
        task_instances_query = task_instances_query.order_by(
            TaskInstance.end_date
        ).limit(quantity)

    results = task_instances_query.all()

    task_instances, dag_runs = _create_task_instances_and_dag_runs(
        results, dagbag, include_xcom, include_logs, session
    )

    return task_instances, dag_runs


@save_result_size("incomplete_task_instances", "incomplete_dag_runs")
@measure_time
def get_incomplete_task_instances_from_completed_dag_runs(
    since, dag_ids, dagbag, max_quantity, offset, session
):
    task_instances_query = (
        session.query(*ETaskInstance.query_fields(), *EDagRun.query_fields())
        .join(
            DagRun,
            (TaskInstance.dag_id == DagRun.dag_id)
            & (TaskInstance.execution_date == DagRun.execution_date),
        )
        .filter(and_(TaskInstance.end_date.is_(None), DagRun.end_date > since))
    )

    if dag_ids:
        task_instances_query = task_instances_query.filter(DagRun.dag_id.in_(dag_ids))

    task_instances_query = (
        task_instances_query.order_by(DagRun.end_date, TaskInstance.task_id)
        .limit(max_quantity)
        .offset(offset)
    )
    results = task_instances_query.all()

    task_instances, dag_runs = _create_task_instances_and_dag_runs(
        results, dagbag, False, False, session
    )

    return task_instances, dag_runs


def _create_task_instances_and_dag_runs(
    results, dagbag, include_xcom, include_logs, session
):
    task_instances = []
    dag_runs = set()
    for fields in results:
        ti_fields = fields[: len(ETaskInstance.db_fields)]
        dr_fields = fields[len(ETaskInstance.db_fields) :]
        task_instances.append(
            build_task_instance(ti_fields, dagbag, include_xcom, include_logs, session)
        )
        dag_runs.add(EDagRun.from_db_fields(*dr_fields))

    return task_instances, dag_runs



import logging

from airflow.models import DagModel
from airflow.utils.db import provide_session
from sqlalchemy.orm import joinedload

from dbnd._core.constants import AD_HOC_DAG_PREFIX
from dbnd_airflow_export.plugin_old.helpers import _get_git_status
from dbnd_airflow_export.plugin_old.metrics import measure_time, save_result_size
from dbnd_airflow_export.plugin_old.model import EDag


logger = logging.getLogger(__name__)


current_dags = {}


@save_result_size("dags")
@measure_time
def get_dags(dagbag, include_task_args, dag_ids, raw_data_only=False):
    dag_models = [d for d in current_dags.values() if d]
    if dag_ids is not None:
        dag_models = [dag for dag in dag_models if dag.dag_id in dag_ids]

    number_of_dags_not_in_dag_bag = 0
    dags_list = []
    git_commit, is_committed = _get_git_status(dagbag.dag_folder)

    for dag_model in dag_models:
        try:
            dag_from_dag_bag = dagbag.get_dag(dag_model.dag_id)
        except Exception:
            logger.debug("DAG %s not in a dagbag", dag_model.dag_id)
            dag_from_dag_bag = None

        if dag_from_dag_bag:
            dag = EDag.from_dag(
                dag_from_dag_bag,
                dag_model,
                dagbag.dag_folder,
                include_task_args,
                git_commit,
                is_committed,
                raw_data_only,
            )
        else:
            dag = EDag.from_dag(
                dag_model,
                dag_model,
                dagbag.dag_folder,
                include_task_args,
                git_commit,
                is_committed,
                raw_data_only,
            )
            number_of_dags_not_in_dag_bag += 1
        dags_list.append(dag)

    if number_of_dags_not_in_dag_bag > 0:
        logger.info("Found %d dags not in dagbag", number_of_dags_not_in_dag_bag)
    return dags_list


def _dag_query(session):
    dag_models = session.query(DagModel)
    if hasattr(DagModel, "tags"):
        # For backward compatibility with AF < 1.10.8
        dag_models = dag_models.options(joinedload(DagModel.tags))
    return dag_models


@save_result_size("current_dags")
@measure_time
def load_dags_models(session):
    dag_models = _dag_query(session)

    for dag_model in dag_models.all():
        # Exclude dbnd-run tagged runs
        if not dag_model.dag_id.startswith(AD_HOC_DAG_PREFIX):
            current_dags[dag_model.dag_id] = dag_model

    logger.info("Collected %d dags", len(current_dags))
    return current_dags


@measure_time
@provide_session
def get_current_dag_model(dag_id, session=None):
    # MONKEY PATCH for old DagModel.get_current to try cache first
    if dag_id not in current_dags:
        current_dags[dag_id] = (
            _dag_query(session).filter(DagModel.dag_id == dag_id).first()
        )

    return current_dags[dag_id]



from datetime import datetime
from typing import List

import attr

from airflow.models import TaskInstance

from dbnd_airflow_export.plugin_old.model import EDag, EDagRun


@attr.s
class AirflowNewDagRun(object):
    id = attr.ib()  # type: int
    dag_id = attr.ib()  # type: str
    execution_date = attr.ib()  # type: datetime
    state = attr.ib()  # type: str
    is_paused = attr.ib()  # type: bool
    has_updated_task_instances = attr.ib()  # type: bool
    max_log_id = attr.ib()  # type: int
    events = attr.ib()  # type: List[str]

    def as_dict(self):
        return dict(
            id=self.id,
            dag_id=self.dag_id,
            execution_date=self.execution_date,
            state=self.state,
            is_paused=self.is_paused,
            has_updated_task_instances=self.has_updated_task_instances,
            max_log_id=self.max_log_id,
            events=self.events,
        )


class AirflowTaskInstance(object):
    def __init__(
        self, dag_id, task_id, execution_date, state, try_number, start_date, end_date,
    ):
        self.execution_date = execution_date
        self.dag_id = dag_id
        self.state = state
        self.try_number = try_number
        self.task_id = task_id
        self.start_date = start_date
        self.end_date = end_date

    db_fields = [
        "dag_id",
        "task_id",
        "execution_date",
        "state",
        "_try_number",
        "start_date",
        "end_date",
    ]

    @classmethod
    def query_fields(cls):
        return [getattr(TaskInstance, key) for key in cls.db_fields]

    def as_dict(self):
        return dict(
            dag_id=self.dag_id,
            task_id=self.task_id,
            execution_date=self.execution_date,
            state=self.state,
            try_number=self.try_number,
            start_date=self.start_date,
            end_date=self.end_date,
        )


@attr.s
class AirflowExportMeta(object):
    airflow_version = attr.ib(default=None)  # type: str
    plugin_version = attr.ib(default=None)  # type: str
    request_args = attr.ib(default=None)  # type: str
    metrics = attr.ib(default=None)  # type: dict

    def as_dict(self):
        return dict(
            airflow_version=self.airflow_version,
            plugin_version=self.plugin_version,
            request_args=self.request_args,
            metrics=self.metrics,
        )


@attr.s
class AirflowExportData(object):
    airflow_export_meta = attr.ib(default=None)  # type: AirflowExportMeta
    error_message = attr.ib(default=None)  # type: str

    def as_dict(self):
        return dict(
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class LastSeenData(AirflowExportData):
    last_seen_dag_run_id = attr.ib(default=None)  # type: int
    last_seen_log_id = attr.ib(default=None)  # type: int

    def as_dict(self):
        return dict(
            last_seen_dag_run_id=self.last_seen_dag_run_id,
            last_seen_log_id=self.last_seen_log_id,
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class NewRunsData(AirflowExportData):
    new_dag_runs = attr.ib(default=None)  # type: List[AirflowNewDagRun]
    last_seen_dag_run_id = attr.ib(default=None)  # type: int
    last_seen_log_id = attr.ib(default=None)  # type: int

    def as_dict(self):
        return dict(
            new_dag_runs=[new_dag_run.as_dict() for new_dag_run in self.new_dag_runs],
            last_seen_dag_run_id=self.last_seen_dag_run_id,
            last_seen_log_id=self.last_seen_log_id,
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class FullRunsData(AirflowExportData):
    task_instances = attr.ib(default=None)  # type: List[AirflowTaskInstance]
    dag_runs = attr.ib(default=None)  # type: List[EDagRun]
    dags = attr.ib(default=None)  # type: List[EDag]

    def as_dict(self):
        return dict(
            task_instances=[
                task_instance.as_dict() for task_instance in self.task_instances
            ],
            dag_runs=[run.as_dict() for run in self.dag_runs],
            dags=[dag.as_dict() for dag in self.dags],
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )


@attr.s
class DagRunsStatesData(AirflowExportData):
    task_instances = attr.ib(default=None)  # type: List[AirflowTaskInstance]
    dag_runs = attr.ib(default=None)  # type: List[EDagRun]

    def as_dict(self):
        return dict(
            task_instances=[
                task_instance.as_dict() for task_instance in self.task_instances
            ],
            dag_runs=[run.as_dict() for run in self.dag_runs],
            airflow_export_meta=self.airflow_export_meta.as_dict(),
            error_message=self.error_message,
        )
